@@ -31,16 +31,17 @@
 # The fact that you are presently reading this means that you have had
 # knowledge of the CeCILL-C license and that you accept its terms.
 #
-# $Id: clush.py 234 2010-02-25 16:52:40Z st-cea $
+# $Id: clush.py 271 2010-06-08 14:29:04Z st-cea $
 
 """
-Utility program to run commands on a cluster using the ClusterShell library.
+Utility program to run commands on a cluster using the ClusterShell
+library.
 
-clush is a pdsh-like command which benefits from the ClusterShell library
-and its Ssh worker. It features an integrated output results gathering
-system (dshbak-like), can get node groups by running predefined external
-commands and can redirect lines read on its standard input to the remote
-commands.
+clush is a pdsh-like command which benefits from the ClusterShell
+library and its Ssh worker. It features an integrated output results
+gathering system (dshbak-like), can get node groups by running
+predefined external commands and can redirect lines read on its
+standard input to the remote commands.
 
 When no command are specified, clush runs interactively.
 
@@ -53,11 +54,22 @@ import sys
 import signal
 import ConfigParser
 
-from ClusterShell.Event import EventHandler
-from ClusterShell.NodeSet import NodeSet, NodeSetParseError
-from ClusterShell.Task import Task, task_self
-from ClusterShell.Worker.Worker import WorkerSimple
-from ClusterShell import __version__
+from ClusterShell.NodeUtils import GroupResolverConfigError
+from ClusterShell.NodeUtils import GroupResolverSourceError
+from ClusterShell.NodeUtils import GroupSourceException
+from ClusterShell.NodeUtils import GroupSourceNoUpcall
+try:
+    from ClusterShell.Event import EventHandler
+    from ClusterShell.NodeSet import NodeSet
+    from ClusterShell.NodeSet import NOGROUP_RESOLVER, STD_GROUP_RESOLVER
+    from ClusterShell.NodeSet import NodeSetExternalError, NodeSetParseError
+    from ClusterShell.Task import Task, task_self
+    from ClusterShell.Worker.Worker import WorkerSimple
+    from ClusterShell import __version__
+except GroupResolverConfigError, e:
+    print >> sys.stderr, \
+        "ERROR: ClusterShell Groups configuration error:\n\t%s" % e
+    sys.exit(1)
 
 VERB_QUIET = 0
 VERB_STD = 1
@@ -114,7 +126,8 @@ class DirectOutputHandler(EventHandler):
             ns = "local"
             rc = worker.retcode()
         if rc > 0:
-            print >> sys.stderr, "clush: %s: exited with exit code %d" % (ns, rc) 
+            print >> sys.stderr, "clush: %s: exited with exit code %d" \
+                                    % (ns, rc) 
 
     def ev_timeout(self, worker):
         print >> sys.stderr, "clush: %s: command timeout" % \
@@ -131,10 +144,29 @@ class DirectOutputHandler(EventHandler):
 class GatherOutputHandler(EventHandler):
     """Gathered output event handler class."""
 
-    def __init__(self, label, gather_print, runtimer):
+    def __init__(self, label, gather_info, runtimer, nodes):
         self._label = label
-        self._gather_print = gather_print
+        self._gather_print, self._regroup, self._groupsource, self._noprefix \
+            = gather_info
         self._runtimer = runtimer
+        # needed for -L enhancement:
+        self._nodes = NodeSet(nodes)
+        self._nodemap_base = dict.fromkeys(self._nodes, 0)
+        self._nodemap = self._nodemap_base.copy()
+        self._nodes_cnt = 0
+
+    def ev_read(self, worker):
+        # Implementation of -bL "per line live gathering":
+        if self._gather_print != print_lines:
+            return
+        result = worker.last_read()
+        if type(result) is not tuple:
+            return
+        # keep counter and nodemap up-to-date
+        node = result[0]
+        self._nodes_cnt += 1
+        self._nodemap[node] += 1
+        self._live_line(worker)
 
     def ev_error(self, worker):
         t = worker.last_error()
@@ -152,6 +184,31 @@ class GatherOutputHandler(EventHandler):
             # Force redisplay of counter
             self._runtimer.eh.set_dirty()
 
+    def ev_hup(self, worker):
+        if self._gather_print == print_lines and \
+                self._nodemap[worker.last_node()] == 0:
+            # forget node that doesn't answer (used for live line gathering)
+            self._nodes.remove(worker.last_node())
+            del self._nodemap[worker.last_node()]
+            del self._nodemap_base[worker.last_node()]
+            self._live_line(worker)
+
+    def _live_line(self, worker):
+        # if all nodes replied 1 (or n) time(s), display gathered line(s)
+        if self._nodes and self._nodes_cnt % len(self._nodes) == 0 and \
+                max(self._nodemap.itervalues()) == (self._nodes_cnt \
+                                                    / len(self._nodes)):
+            nodesetify = lambda v: (v[0], NodeSet.fromlist(v[1]))
+            for buf, nodeset in sorted(map(nodesetify,
+                                           worker.iter_buffers()),
+                                       cmp=bufnodeset_cmp):
+                self._gather_print(nodeset, buf, self._regroup,
+                                   self._groupsource, self._noprefix)
+            # clear worker buffers, reset nodemap and node counter
+            worker.flush_buffers()
+            self._nodemap = self._nodemap_base.copy()
+            self._nodes_cnt = 0
+
     def ev_close(self, worker):
         # Worker is closing -- it's time to gather results...
         if self._runtimer:
@@ -164,13 +221,15 @@ class GatherOutputHandler(EventHandler):
             for buf, nodeset in sorted(map(nodesetify,
                                            worker.iter_buffers(nodelist)),
                                        cmp=bufnodeset_cmp):
-                self._gather_print(nodeset, buf)
+                self._gather_print(nodeset, buf, self._regroup,
+                                   self._groupsource, self._noprefix)
 
         # Display return code if not ok ( != 0)
         for rc, nodelist in worker.iter_retcodes():
             if rc != 0:
                 ns = NodeSet.fromlist(nodelist)
-                print >> sys.stderr, "clush: %s: exited with exit code %s" % (ns, rc)
+                print >> sys.stderr, "clush: %s: exited with exit code %s" \
+                                        % (ns, rc)
 
         # Display nodes that didn't answer within command timeout delay
         if worker.num_timeout() > 0:
@@ -247,7 +306,8 @@ class ClushConfig(ConfigParser.ConfigParser):
         for key, value in ClushConfig.main_defaults.iteritems():
             self.set("Main", key, value)
         # config files override defaults values
-        self.read(['/etc/clustershell/clush.conf', os.path.expanduser('~/.clush.conf')])
+        self.read(['/etc/clustershell/clush.conf',
+                   os.path.expanduser('~/.clush.conf')])
 
     def verbose_print(self, level, message):
         if self.get_verbosity() >= level:
@@ -298,22 +358,6 @@ class ClushConfig(ConfigParser.ConfigParser):
     def get_ssh_options(self):
         return self._get_optional("Main", "ssh_options")
 
-    def get_nodes_all_command(self):
-        section = "External"
-        option = "nodes_all"
-        try:
-            return self.get(section, option)
-        except ConfigParser.Error, e:
-            raise ClushConfigError(section, option, e)
-
-    def get_nodes_group_command(self, group):
-        section = "External"
-        option = "nodes_group"
-        try:
-            return self.get(section, option, 0, { "group" : group })
-        except ConfigParser.Error, e:
-            raise ClushConfigError(section, option, e)
-
 
 def signal_handler(signum, frame):
     """Signal handler used for main thread notification"""
@@ -339,13 +383,21 @@ def readline_setup():
 
 # Start of clubak.py common functions
 
-def print_buffer(header, content):
+def print_buffer(nodeset, content, regroup, groupsource, noprefix):
     """Display a dshbak-like header block and content."""
     sep = "-" * 15
+    if regroup:
+        header = nodeset.regroup(groupsource, noprefix=noprefix)
+    else:
+        header = str(nodeset)
     sys.stdout.write("%s\n%s\n%s\n%s\n" % (sep, header, sep, content))
 
-def print_lines(header, msg):
+def print_lines(nodeset, msg, regroup, groupsource, noprefix):
     """Display a MsgTree buffer by line with prefixed header."""
+    if regroup:
+        header = nodeset.regroup(groupsource, noprefix=noprefix)
+    else:
+        header = str(nodeset)
     for line in msg:
         sys.stdout.write("%s: %s\n" % (header, line))
 
@@ -370,7 +422,8 @@ def bufnodeset_cmp(bn1, bn2):
     # Extract nodesets and call nodeset_cmp
     return nodeset_cmp(bn1[1], bn2[1])
 
-def ttyloop(task, nodeset, gather, timeout, label, verbosity, gather_print):
+def ttyloop(task, nodeset, gather, timeout, label, verbosity,
+            (gather_print, regroup, groupsource, noprefix)):
     """Manage the interactive prompt to run command"""
     readline_avail = False
     if task.default("USER_interactive"):
@@ -421,7 +474,7 @@ def ttyloop(task, nodeset, gather, timeout, label, verbosity, gather_print):
                     if not print_warn:
                         print_warn = True
                         print >> sys.stderr, "Warning: Caught keyboard interrupt!"
-                    gather_print(nodeset, buf)
+                    gather_print(nodeset, buf, regroup, groupsource, noprefix)
                     
                 # Return code handling
                 ns_ok = NodeSet()
@@ -483,14 +536,15 @@ def ttyloop(task, nodeset, gather, timeout, label, verbosity, gather_print):
 
             if cmdl.startswith('!') and len(cmd.strip()) > 0:
                 run_command(task, cmd[1:], None, gather, timeout, None,
-                            verbosity, gather_print)
+                            verbosity, (gather_print, regroup, groupsource,
+                                        noprefix))
             elif cmdl != "quit":
                 if not cmd:
                     continue
                 if readline_avail:
                     readline.write_history_file(get_history_file())
                 run_command(task, cmd, ns, gather, timeout, label, verbosity,
-                            gather_print)
+                            (gather_print, regroup, groupsource, noprefix))
     return rc
 
 def bind_stdin(worker):
@@ -508,7 +562,7 @@ def bind_stdin(worker):
     # Add stdin worker to the same task than given worker
     worker.task.schedule(worker_stdin)
 
-def run_command(task, cmd, ns, gather, timeout, label, verbosity, gather_print):
+def run_command(task, cmd, ns, gather, timeout, label, verbosity, gather_info):
     """
     Create and run the specified command line, displaying
     results in a dshbak way when gathering is used.
@@ -523,7 +577,7 @@ def run_command(task, cmd, ns, gather, timeout, label, verbosity, gather_print):
             runtimer = task.timer(2.0, RunTimer(task, len(ns)), interval=1./3.,
                                   autoclose=True)
         worker = task.shell(cmd, nodes=ns, handler=GatherOutputHandler(label,
-                            gather_print, runtimer), timeout=timeout)
+                            gather_info, runtimer, ns), timeout=timeout)
     else:
         worker = task.shell(cmd, nodes=ns, handler=DirectOutputHandler(label),
                             timeout=timeout)
@@ -596,16 +650,27 @@ def clush_main(args):
     optgrp.add_option("-d", "--debug", action="store_true", dest="debug",
                       help="output more messages for debugging purpose")
 
-    optgrp.add_option("-L", action="store_true", dest="line_mode", default=False,
-                      help="disable header block and order output by nodes")
+    optgrp.add_option("-G", "--groupbase", action="store_true",
+                      dest="groupbase", default=False, help="do not display " \
+                      "group source prefix")
+    optgrp.add_option("-L", action="store_true", dest="line_mode",
+                      default=False, help="disable header block and order " \
+                      "output by nodes")
     optgrp.add_option("-N", action="store_false", dest="label", default=True,
                       help="disable labeling of command line")
     optgrp.add_option("-S", action="store_true", dest="maxrc",
                       help="return the largest of command return codes")
     optgrp.add_option("-b", "--dshbak", action="store_true", dest="gather",
-                      default=False, help="display gathered results in a dshbak-like way")
+                      default=False, help="display gathered results in a " \
+                      "dshbak-like way")
     optgrp.add_option("-B", action="store_true", dest="gatherall",
-                      default=False, help="like -b but including standard error")
+                      default=False, help="like -b but including standard " \
+                      "error")
+    optgrp.add_option("-r", "--regroup", action="store_true", dest="regroup",
+                      default=False, help="fold nodeset using node groups")
+    optgrp.add_option("-s", "--groupsource", action="store",
+                      dest="groupsource", help="optional groups.conf(5) " \
+                      "group source to use")
     parser.add_option_group(optgrp)
 
     # Copy
@@ -626,8 +691,9 @@ def clush_main(args):
                       help="execute remote command as user")
     optgrp.add_option("-o", "--options", action="store", dest="options",
                       help="can be used to give ssh options")
-    optgrp.add_option("-t", "--connect_timeout", action="store", dest="connect_timeout", 
-                      help="limit time to connect to a node" ,type="float")
+    optgrp.add_option("-t", "--connect_timeout", action="store",
+                      dest="connect_timeout", help="limit time to connect to " \
+                      "a node" ,type="float")
     optgrp.add_option("-u", "--command_timeout", action="store", dest="command_timeout", 
                       help="limit time for command to run on the node", type="float")
     parser.add_option_group(optgrp)
@@ -665,32 +731,38 @@ def clush_main(args):
     if options.exclude:
         nodeset_exclude = NodeSet.fromlist(options.exclude)
 
+    if options.groupsource:
+        # Be sure -a/g -s source work as espected.
+        STD_GROUP_RESOLVER.default_sourcename = options.groupsource
+
     # Do we have nodes group?
     task = task_self()
     task.set_info("debug", config.get_verbosity() > 1)
+    if config.get_verbosity() > 1:
+        STD_GROUP_RESOLVER.set_verbosity(1)
     if options.nodes_all:
-        command = config.get_nodes_all_command()
-        task.shell(command, key="all")
+        all_nodeset = NodeSet.fromall()
+        config.verbose_print(VERB_DEBUG, \
+            "Adding nodes from option -a: %s" % all_nodeset)
+        nodeset_base.add(all_nodeset)
+
     if options.group:
-        for grp in options.group:
-            command = config.get_nodes_group_command(grp)
-            task.shell(command, key="group")
+        grp_nodeset = NodeSet.fromlist(options.group,
+                                       resolver=NOGROUP_RESOLVER)
+        for grp in grp_nodeset:
+            addingrp = NodeSet("@" + grp)
+            config.verbose_print(VERB_DEBUG, \
+                "Adding nodes from option -g %s: %s" % (grp, addingrp))
+            nodeset_base.update(addingrp)
+
     if options.exgroup:
-        for grp in options.exgroup:
-            command = config.get_nodes_group_command(grp)
-            task.shell(command, key="exgroup")
-
-    # Run needed external commands
-    task.resume()
-
-    for buf, keys in task.iter_buffers(['all', 'group']):
-        for line in buf:
-            config.verbose_print(VERB_DEBUG, "Adding nodes from option %s: %s" % (','.join(keys), buf))
-            nodeset_base.add(line)
-    for buf, keys in task.iter_buffers(['exgroup']):
-        for line in buf:
-            config.verbose_print(VERB_DEBUG, "Excluding nodes from option %s: %s" % (','.join(keys), buf))
-            nodeset_exclude.add(line)
+        grp_nodeset = NodeSet.fromlist(options.exgroup,
+                                       resolver=NOGROUP_RESOLVER)
+        for grp in grp_nodeset:
+            removingrp = NodeSet("@" + grp)
+            config.verbose_print(VERB_DEBUG, \
+                "Excluding nodes from option -X %s: %s" % (grp, removingrp))
+            nodeset_exclude.update(removingrp)
 
     # Do we have an exclude list? (-x ...)
     nodeset_base.difference_update(nodeset_exclude)
@@ -780,9 +852,11 @@ def clush_main(args):
 
     # Select gather-mode print function
     if options.line_mode:
-        gather_print = print_lines
+        gather_info = print_lines, options.regroup, options.groupsource, \
+                        options.groupbase
     else:
-        gather_print = print_buffer
+        gather_info = print_buffer, options.regroup, options.groupsource, \
+                        options.groupbase
 
     if not task.default("USER_interactive"):
         if options.source_path:
@@ -790,11 +864,11 @@ def clush_main(args):
                     0, options.preserve_flag)
         else:
             run_command(task, ' '.join(args), nodeset_base, gather, timeout,
-                        options.label, config.get_verbosity(), gather_print)
+                        options.label, config.get_verbosity(), gather_info)
 
     if user_interaction:
         ttyloop(task, nodeset_base, gather, timeout, options.label,
-                config.get_verbosity(), gather_print)
+                config.get_verbosity(), gather_info)
     elif task.default("USER_interactive"):
         print >> sys.stderr, "ERROR: interactive mode requires a tty"
         clush_exit(1)
@@ -813,8 +887,21 @@ if __name__ == '__main__':
     except ClushConfigError, e:
         print >> sys.stderr, "ERROR: %s" % e
         sys.exit(1)
+    except NodeSetExternalError, e:
+        print >> sys.stderr, "clush: external error:", e
+        sys.exit(1)
     except NodeSetParseError, e:
-        print >> sys.stderr, "NodeSet parse error:", e
+        print >> sys.stderr, "clush: parse error:", e
+        sys.exit(1)
+    except GroupResolverSourceError, e:
+        print >> sys.stderr, "ERROR: unknown group source: \"%s\"" % e
+        sys.exit(1)
+    except GroupSourceNoUpcall, e:
+        print >> sys.stderr, "ERROR: no %s upcall defined for group " \
+            "source \"%s\"" % (e, e.group_source.name)
+        sys.exit(1)
+    except GroupSourceException, e:
+        print >> sys.stderr, "ERROR: other group error:", e
         sys.exit(1)
     except IOError:
         # Ignore broken pipe
